@@ -4,99 +4,61 @@
 
 # Number of CPUs (for make -j)
 NCPU=`sysctl -n hw.ncpu`
-NJOB=$NCPU
-#NJOB=`expr $NCPU + 1`
+if test x$NJOB = x; then
+    NJOB=$NCPU
+fi
+
+# SDK path
+if test x$SDK_PATH = x; then
+    SDK_PATH=/Developer/SDKs
+fi
 
 # Generic, cross-platform CFLAGS you always want go here.
 CFLAGS="-O3 -g -pipe"
 
-# Locate Xcode SDK path
-SDK_PATH=/Developer/SDKs
-if [ ! -d $SDK_PATH ]; then
-    echo "Couldn't find SDK path"
-    exit 1
-fi
-
-# See if we can use 10.2 or 10.3 runtime compatibility
-if [ -d "$SDK_PATH/MacOSX10.2.8.sdk" ]; then
-    # PowerPC configure flags (10.2 runtime compatibility)
-    # We dynamically load X11, so using the system X11 headers is fine.
-    CONFIG_PPC="--build=`uname -p`-apple-darwin --host=powerpc-apple-darwin \
---x-includes=/usr/X11R6/include --x-libraries=/usr/X11R6/lib"
-
-    # PowerPC compiler flags
-    CC_PPC="gcc-3.3 -arch ppc"
-    CXX_PPC="g++-3.3 -arch ppc"
-    CFLAGS_PPC=""
-    CPPFLAGS_PPC="-DMAC_OS_X_VERSION_MIN_REQUIRED=1020 \
--nostdinc \
--F$SDK_PATH/MacOSX10.2.8.sdk/System/Library/Frameworks \
--I$SDK_PATH/MacOSX10.2.8.sdk/usr/include/gcc/darwin/3.3 \
--isystem $SDK_PATH/MacOSX10.2.8.sdk/usr/include"
-
-    # PowerPC linker flags 
-    LFLAGS_PPC="-arch ppc \
--L$SDK_PATH/MacOSX10.2.8.sdk/usr/lib/gcc/darwin/3.3 \
--F$SDK_PATH/MacOSX10.2.8.sdk/System/Library/Frameworks \
--Wl,-syslibroot,$SDK_PATH/MacOSX10.2.8.sdk"
-
-else # 10.2 or 10.3 SDK
-
-    # PowerPC configure flags (10.3 runtime compatibility)
-    # We dynamically load X11, so using the system X11 headers is fine.
-    CONFIG_PPC="--build=`uname -p`-apple-darwin --host=powerpc-apple-darwin \
---x-includes=/usr/X11R6/include --x-libraries=/usr/X11R6/lib"
-
-    # PowerPC compiler flags
-    CC_PPC="gcc-4.0 -arch ppc"
-    CXX_PPC="g++-4.0 -arch ppc"
-    CFLAGS_PPC=""
-    CPPFLAGS_PPC="-DMAC_OS_X_VERSION_MIN_REQUIRED=1030 \
--nostdinc \
--F$SDK_PATH/MacOSX10.3.9.sdk/System/Library/Frameworks \
--I$SDK_PATH/MacOSX10.3.9.sdk/usr/lib/gcc/powerpc-apple-darwin9/4.0.1/include \
--isystem $SDK_PATH/MacOSX10.3.9.sdk/usr/include"
-
-    # PowerPC linker flags
-    LFLAGS_PPC="-arch ppc -mmacosx-version-min=10.3 \
--L$SDK_PATH/MacOSX10.3.9.sdk/usr/lib/gcc/powerpc-apple-darwin9/4.0.1 \
--F$SDK_PATH/MacOSX10.3.9.sdk/System/Library/Frameworks \
--Wl,-syslibroot,$SDK_PATH/MacOSX10.3.9.sdk"
-
-fi # 10.2 or 10.3 SDK
-
-# Intel configure flags (10.4 runtime compatibility)
 # We dynamically load X11, so using the system X11 headers is fine.
-CONFIG_X86="--build=`uname -p`-apple-darwin --host=i386-apple-darwin \
+BASE_CONFIG_FLAGS="--build=`uname -p`-apple-darwin \
 --x-includes=/usr/X11R6/include --x-libraries=/usr/X11R6/lib"
 
-# Intel compiler flags
-CC_X86="gcc-4.0 -arch i386"
-CXX_X86="g++-4.0 -arch i386"
-CFLAGS_X86="-mmacosx-version-min=10.4"
-CPPFLAGS_X86="-DMAC_OS_X_VERSION_MIN_REQUIRED=1040 \
--nostdinc \
--F$SDK_PATH/MacOSX10.4u.sdk/System/Library/Frameworks \
--I$SDK_PATH/MacOSX10.4u.sdk/usr/lib/gcc/i686-apple-darwin9/4.0.1/include \
--isystem $SDK_PATH/MacOSX10.4u.sdk/usr/include"
+# PowerPC 32-bit compiler flags
+CONFIG_PPC="--host=powerpc-apple-darwin"
+CC_PPC="gcc-4.0"
+CXX_PPC="g++-4.0"
+BUILD_FLAGS_PPC="-arch ppc -mmacosx-version-min=10.4"
 
-# Intel linker flags
-LFLAGS_X86="-arch i386 -mmacosx-version-min=10.4 \
--L$SDK_PATH/MacOSX10.4u.sdk/usr/lib/gcc/i686-apple-darwin9/4.0.1 \
--Wl,-syslibroot,$SDK_PATH/MacOSX10.4u.sdk"
+# Intel 32-bit compiler flags
+CONFIG_X86="--host=i386-apple-darwin"
+CC_X86="gcc"
+CXX_X86="g++"
+BUILD_FLAGS_X86="-arch i386 -mmacosx-version-min=10.4"
+
+# Intel 64-bit compiler flags
+CONFIG_X64="--host=x86_64-apple-darwin"
+CC_X64="gcc"
+CXX_X64="g++"
+BUILD_FLAGS_X64="-arch x86_64 -mmacosx-version-min=10.6"
 
 #
 # Find the configure script
 #
 srcdir=`dirname $0`/..
+srcdir=`cd $srcdir && pwd`
 auxdir=$srcdir/build-scripts
 cd $srcdir
 
+allow_ppc="yes"
+which gcc-4.0 >/dev/null 2>/dev/null
+if [ "x$?" = "x1" ]; then
+    #echo "WARNING: Can't find gcc-4.0, which means you don't have Xcode 3."
+    #echo "WARNING: Therefore, we can't do PowerPC support."
+    allow_ppc="no"
+fi
+
 #
 # Figure out which phase to build:
 # all,
-# configure, configure-ppc, configure-x86,
-# make, make-ppc, make-x86, merge
+# configure, configure-ppc, configure-x86, configure-x64
+# make, make-ppc, make-x86, make-x64, merge
 # install
 # clean
 if test x"$1" = x; then
@@ -106,33 +68,43 @@ else
 fi
 case $phase in
     all)
-        configure_ppc="yes"
+        configure_ppc="$allow_ppc"
         configure_x86="yes"
-        make_ppc="yes"
+        configure_x64="yes"
+        make_ppc="$allow_ppc"
         make_x86="yes"
+        make_x64="yes"
         merge="yes"
         ;;
     configure)
-        configure_ppc="yes"
+        configure_ppc="$allow_ppc"
         configure_x86="yes"
+        configure_x64="yes"
         ;;
     configure-ppc)
-        configure_ppc="yes"
+        configure_ppc="$allow_ppc"
         ;;
     configure-x86)
         configure_x86="yes"
         ;;
+    configure-x64)
+        configure_x64="yes"
+        ;;
     make)
-        make_ppc="yes"
+        make_ppc="$allow_ppc"
         make_x86="yes"
+        make_x64="yes"
         merge="yes"
         ;;
     make-ppc)
-        make_ppc="yes"
+        make_ppc="$allow_ppc"
         ;;
     make-x86)
         make_x86="yes"
         ;;
+    make-x64)
+        make_x64="yes"
+        ;;
     merge)
         merge="yes"
         ;;
@@ -161,6 +133,7 @@ case $phase in
     clean)
         clean_ppc="yes"
         clean_x86="yes"
+        clean_x64="yes"
         ;;
     clean-ppc)
         clean_ppc="yes"
@@ -168,18 +141,24 @@ case $phase in
     clean-x86)
         clean_x86="yes"
         ;;
+    clean-x64)
+        clean_x64="yes"
+        ;;
     *)
-        echo "Usage: $0 [all|configure[-ppc|-x86]|make[-ppc|-x86]|merge|install|clean]"
+        echo "Usage: $0 [all|configure[-ppc|-x86|-x64]|make[-ppc|-x86|-x64]|merge|install|clean[-ppc|-x86|-x64]]"
         exit 1
         ;;
 esac
 case `uname -p` in
-    powerpc)
-        native_path=ppc
-        ;;
     *86)
         native_path=x86
         ;;
+    *powerpc)
+        native_path=ppc
+        ;;
+    x86_64)
+        native_path=x64
+        ;;
     *)
         echo "Couldn't figure out native architecture path"
         exit 1
@@ -189,7 +168,7 @@ esac
 #
 # Create the build directories
 #
-for dir in build build/ppc build/x86; do
+for dir in build build/ppc build/x86 build/x64; do
     if test -d $dir; then
         :
     else
@@ -197,28 +176,39 @@ for dir in build build/ppc build/x86; do
     fi
 done
 
+
 #
-# Build the PowerPC binary
+# Build the PowerPC 32-bit binary
 #
 if test x$configure_ppc = xyes; then
     (cd build/ppc && \
-     sh ../../configure $CONFIG_PPC CC="$CC_PPC" CXX="$CXX_PPC" CFLAGS="$CFLAGS $CFLAGS_PPC" CPPFLAGS="$CPPFLAGS_PPC" LDFLAGS="$LFLAGS_PPC") || exit 2
+     sh ../../configure $BASE_CONFIG_FLAGS $CONFIG_PPC CC="$CC_PPC" CXX="$CXX_PPC" CFLAGS="$CFLAGS $BUILD_FLAGS_PPC $CFLAGS_PPC" LDFLAGS="$BUILD_FLAGS_PPC $LFLAGS_PPC") || exit 2
 fi
 if test x$make_ppc = xyes; then
-    (cd build/ppc && ls include && make -j$NJOB) || exit 3
+    (cd build/ppc && make -j$NJOB) || exit 3
 fi
-
 #
-# Build the Intel binary
+# Build the Intel 32-bit binary
 #
 if test x$configure_x86 = xyes; then
     (cd build/x86 && \
-     sh ../../configure $CONFIG_X86 CC="$CC_X86" CXX="$CXX_X86" CFLAGS="$CFLAGS $CFLAGS_X86" CPPFLAGS="$CPPFLAGS_X86" LDFLAGS="$LFLAGS_X86") || exit 2
+     sh ../../configure $BASE_CONFIG_FLAGS $CONFIG_X86 CC="$CC_X86" CXX="$CXX_X86" CFLAGS="$CFLAGS $BUILD_FLAGS_X86 $CFLAGS_X86" LDFLAGS="$BUILD_FLAGS_X86 $LFLAGS_X86") || exit 2
 fi
 if test x$make_x86 = xyes; then
     (cd build/x86 && make -j$NJOB) || exit 3
 fi
 
+#
+# Build the Intel 64-bit binary
+#
+if test x$configure_x64 = xyes; then
+    (cd build/x64 && \
+     sh ../../configure $BASE_CONFIG_FLAGS $CONFIG_X64 CC="$CC_X64" CXX="$CXX_X64" CFLAGS="$CFLAGS $BUILD_FLAGS_X64 $CFLAGS_X64" LDFLAGS="$BUILD_FLAGS_X64 $LFLAGS_X64") || exit 2
+fi
+if test x$make_x64 = xyes; then
+    (cd build/x64 && make -j$NJOB) || exit 3
+fi
+
 #
 # Combine into fat binary
 #
@@ -226,15 +216,17 @@ if test x$merge = xyes; then
     output=.libs
     sh $auxdir/mkinstalldirs build/$output
     cd build
-    target=`find . -mindepth 3 -type f -name '*.dylib' | head -1 | sed 's|.*/||'`
-    (lipo -create -o $output/$target `find . -mindepth 3 -type f -name "*.dylib"` &&
-     ln -sf $target $output/libSDL-1.2.0.dylib &&
+    target=`find . -mindepth 4 -maxdepth 4 -type f -name '*.dylib' | head -1 | sed 's|.*/||'`
+    (lipo -create -o $output/$target `find . -mindepth 4 -maxdepth 4 -type f -name "*.dylib"` &&
      ln -sf $target $output/libSDL.dylib &&
      lipo -create -o $output/libSDL.a */build/.libs/libSDL.a &&
      cp $native_path/build/.libs/libSDL.la $output &&
      cp $native_path/build/.libs/libSDL.lai $output &&
      cp $native_path/build/libSDL.la . &&
-     lipo -create -o libSDLmain.a */build/libSDLmain.a &&
+     lipo -create -o $output/libSDLmain.a */build/.libs/libSDLmain.a &&
+     cp $native_path/build/.libs/libSDLmain.la $output &&
+     cp $native_path/build/.libs/libSDLmain.lai $output &&
+     cp $native_path/build/libSDLmain.la . &&
      echo "Build complete!" &&
      echo "Files can be found in the build directory.") || exit 4
     cd ..
@@ -284,12 +276,13 @@ fi
 if test x$install_lib = xyes; then
     do_install sh $auxdir/mkinstalldirs $libdir
     do_install sh build/$native_path/libtool --mode=install /usr/bin/install -c  build/libSDL.la $libdir/libSDL.la
-    do_install /usr/bin/install -c -m 644 build/libSDLmain.a $libdir/libSDLmain.a
-    do_install ranlib $libdir/libSDLmain.a
+    do_install sh build/$native_path/libtool --mode=install /usr/bin/install -c  build/libSDLmain.la $libdir/libSDLmain.la
 fi
 if test x$install_data = xyes; then
     do_install sh $auxdir/mkinstalldirs $datadir/aclocal
     do_install /usr/bin/install -c -m 644 $srcdir/sdl.m4 $datadir/aclocal/sdl.m4
+    do_install sh $auxdir/mkinstalldirs $libdir/pkgconfig
+    do_install /usr/bin/install -m 644 build/$native_path/sdl.pc $libdir/pkgconfig/sdl.pc
 fi
 if test x$install_man = xyes; then
     do_install sh $auxdir/mkinstalldirs $mandir/man3
@@ -307,10 +300,12 @@ do_clean()
     echo $*
     $* || exit 6
 }
-if test x$clean_x86 = xyes; then
-    do_clean rm -r build/x86
-fi
 if test x$clean_ppc = xyes; then
     do_clean rm -r build/ppc
 fi
-
+if test x$clean_x86 = xyes; then
+    do_clean rm -r build/x86
+fi
+if test x$clean_x64 = xyes; then
+    do_clean rm -r build/x64
+fi
