@@ -1,6 +1,6 @@
 /*
     SDL - Simple DirectMedia Layer
-    Copyright (C) 1997-2006 Sam Lantinga
+    Copyright (C) 1997-2012 Sam Lantinga
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -26,8 +26,12 @@
 #include "SDL.h"
 #include "SDL_cpuinfo.h"
 
-#if defined(__MACOSX__) && defined(__ppc__)
+#if defined(__MACOSX__) && (defined(__ppc__) || defined(__ppc64__))
 #include <sys/sysctl.h> /* For AltiVec check */
+#elif defined(__OpenBSD__) && defined(__powerpc__)
+#include <sys/param.h>
+#include <sys/sysctl.h> /* For AltiVec check */
+#include <machine/cpu.h>
 #elif SDL_ALTIVEC_BLITTERS && HAVE_SETJMP
 #include <signal.h>
 #include <setjmp.h>
@@ -41,8 +45,10 @@
 #define CPU_HAS_SSE	0x00000040
 #define CPU_HAS_SSE2	0x00000080
 #define CPU_HAS_ALTIVEC	0x00000100
+#define CPU_HAS_ARM_SIMD 0x00000200
+#define CPU_HAS_ARM_NEON 0x00000400
 
-#if SDL_ALTIVEC_BLITTERS && HAVE_SETJMP && !__MACOSX__
+#if SDL_ALTIVEC_BLITTERS && HAVE_SETJMP && !__MACOSX__ && !__OpenBSD__
 /* This is the brute force way of detecting instruction sets...
    the idea is borrowed from the libmpeg2 library - thanks!
  */
@@ -146,52 +152,80 @@ done:
 static __inline__ int CPU_getCPUIDFeatures(void)
 {
 	int features = 0;
-#if defined(__GNUC__) && ( defined(i386) || defined(__x86_64__) )
+#if defined(__GNUC__) && defined(i386)
 	__asm__ (
-"        movl    %%ebx,%%edi\n"
 "        xorl    %%eax,%%eax         # Set up for CPUID instruction    \n"
+"        pushl   %%ebx                                                 \n"
 "        cpuid                       # Get and save vendor ID          \n"
+"        popl    %%ebx                                                 \n"
 "        cmpl    $1,%%eax            # Make sure 1 is valid input for CPUID\n"
 "        jl      1f                  # We dont have the CPUID instruction\n"
 "        xorl    %%eax,%%eax                                           \n"
 "        incl    %%eax                                                 \n"
+"        pushl   %%ebx                                                 \n"
 "        cpuid                       # Get family/model/stepping/features\n"
+"        popl    %%ebx                                                 \n"
 "        movl    %%edx,%0                                              \n"
 "1:                                                                    \n"
-"        movl    %%edi,%%ebx\n"
 	: "=m" (features)
 	:
-	: "%eax", "%ecx", "%edx", "%edi"
+	: "%eax", "%ecx", "%edx"
+	);
+#elif defined(__GNUC__) && defined(__x86_64__)
+	__asm__ (
+"        xorl    %%eax,%%eax         # Set up for CPUID instruction    \n"
+"        pushq   %%rbx                                                 \n"
+"        cpuid                       # Get and save vendor ID          \n"
+"        popq    %%rbx                                                 \n"
+"        cmpl    $1,%%eax            # Make sure 1 is valid input for CPUID\n"
+"        jl      1f                  # We dont have the CPUID instruction\n"
+"        xorl    %%eax,%%eax                                           \n"
+"        incl    %%eax                                                 \n"
+"        pushq   %%rbx                                                 \n"
+"        cpuid                       # Get family/model/stepping/features\n"
+"        popq    %%rbx                                                 \n"
+"        movl    %%edx,%0                                              \n"
+"1:                                                                    \n"
+	: "=m" (features)
+	:
+	: "%rax", "%rcx", "%rdx"
 	);
 #elif (defined(_MSC_VER) && defined(_M_IX86)) || defined(__WATCOMC__)
 	__asm {
         xor     eax, eax            ; Set up for CPUID instruction
+        push    ebx
         cpuid                       ; Get and save vendor ID
+        pop     ebx
         cmp     eax, 1              ; Make sure 1 is valid input for CPUID
         jl      done                ; We dont have the CPUID instruction
         xor     eax, eax
         inc     eax
+        push    ebx
         cpuid                       ; Get family/model/stepping/features
+        pop     ebx
         mov     features, edx
 done:
 	}
 #elif defined(__sun) && (defined(__i386) || defined(__amd64))
 	    __asm(
-"        movl    %ebx,%edi\n"
 "        xorl    %eax,%eax         \n"
+"        pushl   %ebx              \n"
 "        cpuid                     \n"
+"        popl    %ebx              \n"
 "        cmpl    $1,%eax           \n"
 "        jl      1f                \n"
 "        xorl    %eax,%eax         \n"
 "        incl    %eax              \n"
+"        pushl   %ebx              \n"
 "        cpuid                     \n"
+"        popl    %ebx              \n"
 #ifdef __i386
 "        movl    %edx,-8(%ebp)     \n"
 #else
 "        movl    %edx,-8(%rbp)     \n"
 #endif
 "1:                                \n"
-"        movl    %edi,%ebx\n" );
+	    );
 #endif
 	return features;
 }
@@ -199,49 +233,75 @@ done:
 static __inline__ int CPU_getCPUIDFeaturesExt(void)
 {
 	int features = 0;
-#if defined(__GNUC__) && (defined(i386) || defined (__x86_64__) )
+#if defined(__GNUC__) && defined(i386)
 	__asm__ (
-"        movl    %%ebx,%%edi\n"
 "        movl    $0x80000000,%%eax   # Query for extended functions    \n"
+"        pushl   %%ebx                                                 \n"
 "        cpuid                       # Get extended function limit     \n"
+"        popl    %%ebx                                                 \n"
 "        cmpl    $0x80000001,%%eax                                     \n"
 "        jl      1f                  # Nope, we dont have function 800000001h\n"
 "        movl    $0x80000001,%%eax   # Setup extended function 800000001h\n"
+"        pushl   %%ebx                                                 \n"
 "        cpuid                       # and get the information         \n"
+"        popl    %%ebx                                                 \n"
 "        movl    %%edx,%0                                              \n"
 "1:                                                                    \n"
-"        movl    %%edi,%%ebx\n"
 	: "=m" (features)
 	:
-	: "%eax", "%ecx", "%edx", "%edi"
+	: "%eax", "%ecx", "%edx"
+	);
+#elif defined(__GNUC__) && defined (__x86_64__)
+	__asm__ (
+"        movl    $0x80000000,%%eax   # Query for extended functions    \n"
+"        pushq   %%rbx                                                 \n"
+"        cpuid                       # Get extended function limit     \n"
+"        popq    %%rbx                                                 \n"
+"        cmpl    $0x80000001,%%eax                                     \n"
+"        jl      1f                  # Nope, we dont have function 800000001h\n"
+"        movl    $0x80000001,%%eax   # Setup extended function 800000001h\n"
+"        pushq   %%rbx                                                 \n"
+"        cpuid                       # and get the information         \n"
+"        popq    %%rbx                                                 \n"
+"        movl    %%edx,%0                                              \n"
+"1:                                                                    \n"
+	: "=m" (features)
+	:
+	: "%rax", "%rcx", "%rdx"
 	);
 #elif (defined(_MSC_VER) && defined(_M_IX86)) || defined(__WATCOMC__)
 	__asm {
         mov     eax,80000000h       ; Query for extended functions
+        push    ebx
         cpuid                       ; Get extended function limit
+        pop     ebx
         cmp     eax,80000001h
         jl      done                ; Nope, we dont have function 800000001h
         mov     eax,80000001h       ; Setup extended function 800000001h
+        push    ebx
         cpuid                       ; and get the information
+        pop     ebx
         mov     features,edx
 done:
 	}
 #elif defined(__sun) && ( defined(__i386) || defined(__amd64) )
 	    __asm (
-"        movl    %ebx,%edi\n"
 "        movl    $0x80000000,%eax \n"
+"        pushl   %ebx             \n"
 "        cpuid                    \n"
+"        popl    %ebx             \n"
 "        cmpl    $0x80000001,%eax \n"
 "        jl      1f               \n"
 "        movl    $0x80000001,%eax \n"
+"        pushl   %ebx             \n"
 "        cpuid                    \n"
+"        popl    %ebx             \n"
 #ifdef __i386
 "        movl    %edx,-8(%ebp)   \n"
 #else
 "        movl    %edx,-8(%rbp)   \n"
 #endif
 "1:                               \n"
-"        movl    %edi,%ebx\n"
 	    );
 #endif
 	return features;
@@ -306,8 +366,12 @@ static __inline__ int CPU_haveSSE2(void)
 static __inline__ int CPU_haveAltiVec(void)
 {
 	volatile int altivec = 0;
-#if defined(__MACOSX__) && defined(__ppc__)
-	int selectors[2] = { CTL_HW, HW_VECTORUNIT }; 
+#if (defined(__MACOSX__) && (defined(__ppc__) || defined(__ppc64__))) || (defined(__OpenBSD__) && defined(__powerpc__))
+# ifdef __OpenBSD__
+	int selectors[2] = { CTL_MACHDEP, CPU_ALTIVEC };
+# else
+	int selectors[2] = { CTL_HW, HW_VECTORUNIT };
+# endif
 	int hasVectorUnit = 0; 
 	size_t length = sizeof(hasVectorUnit); 
 	int error = sysctl(selectors, 2, &hasVectorUnit, &length, NULL, 0); 
@@ -328,6 +392,106 @@ static __inline__ int CPU_haveAltiVec(void)
 	return altivec; 
 }
 
+#ifdef __linux__
+
+#include <unistd.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <fcntl.h>
+#include <elf.h>
+
+static __inline__ int CPU_haveARMSIMD(void)
+{
+	int arm_simd = 0;
+	int fd;
+
+	fd = open("/proc/self/auxv", O_RDONLY);
+	if (fd >= 0)
+	{
+		Elf32_auxv_t aux;
+		while (read(fd, &aux, sizeof aux) == sizeof aux)
+		{
+			if (aux.a_type == AT_PLATFORM)
+			{
+				const char *plat = (const char *) aux.a_un.a_val;
+				arm_simd = strncmp(plat, "v6l", 3) == 0 ||
+				           strncmp(plat, "v7l", 3) == 0;
+			}
+		}
+		close(fd);
+	}
+	return arm_simd;
+}
+
+static __inline__ int CPU_haveARMNEON(void)
+{
+	int arm_neon = 0;
+	int fd;
+
+	fd = open("/proc/self/auxv", O_RDONLY);
+	if (fd >= 0)
+	{
+		Elf32_auxv_t aux;
+		while (read(fd, &aux, sizeof aux) == sizeof aux)
+		{
+			if (aux.a_type == AT_HWCAP)
+				arm_neon = (aux.a_un.a_val & 4096) != 0;
+		}
+		close(fd);
+	}
+	return arm_neon;
+}
+
+#elif defined(__RISCOS__)
+
+#include <kernel.h>
+#include <swis.h>
+
+static __inline__ int CPU_haveARMSIMD(void)
+{
+	_kernel_swi_regs regs;
+	regs.r[0] = 0;
+	if (_kernel_swi(OS_PlatformFeatures, &regs, &regs) != NULL)
+		return 0;
+
+	if (!(regs.r[0] & (1<<31)))
+		return 0;
+
+	regs.r[0] = 34;
+	regs.r[1] = 29;
+	if (_kernel_swi(OS_PlatformFeatures, &regs, &regs) != NULL)
+		return 0;
+
+	return regs.r[0];
+}
+
+static __inline__ int CPU_haveARMNEON(void)
+{
+	/* Use the VFPSupport_Features SWI to access the MVFR registers */
+	_kernel_swi_regs regs;
+	regs.r[0] = 0;
+	if (_kernel_swi(VFPSupport_Features, &regs, &regs) == NULL) {
+		if ((regs.r[2] & 0xFFF000) == 0x111000) {
+			return 1;
+		}
+	}
+	return 0;
+}
+
+#else
+
+static __inline__ int CPU_haveARMSIMD(void)
+{
+	return 0;
+}
+
+static __inline__ int CPU_haveARMNEON(void)
+{
+	return 0;
+}
+
+#endif
+
 static Uint32 SDL_CPUFeatures = 0xFFFFFFFF;
 
 static Uint32 SDL_GetCPUFeatures(void)
@@ -358,6 +522,12 @@ static Uint32 SDL_GetCPUFeatures(void)
 		if ( CPU_haveAltiVec() ) {
 			SDL_CPUFeatures |= CPU_HAS_ALTIVEC;
 		}
+		if ( CPU_haveARMSIMD() ) {
+			SDL_CPUFeatures |= CPU_HAS_ARM_SIMD;
+		}
+		if ( CPU_haveARMNEON() ) {
+			SDL_CPUFeatures |= CPU_HAS_ARM_NEON;
+		}
 	}
 	return SDL_CPUFeatures;
 }
@@ -426,6 +596,22 @@ SDL_bool SDL_HasAltiVec(void)
 	return SDL_FALSE;
 }
 
+SDL_bool SDL_HasARMSIMD(void)
+{
+	if ( SDL_GetCPUFeatures() & CPU_HAS_ARM_SIMD ) {
+		return SDL_TRUE;
+	}
+	return SDL_FALSE;
+}
+
+SDL_bool SDL_HasARMNEON(void)
+{
+	if ( SDL_GetCPUFeatures() & CPU_HAS_ARM_NEON ) {
+		return SDL_TRUE;
+	}
+	return SDL_FALSE;
+}
+
 #ifdef TEST_MAIN
 
 #include <stdio.h>
@@ -440,6 +626,8 @@ int main()
 	printf("SSE: %d\n", SDL_HasSSE());
 	printf("SSE2: %d\n", SDL_HasSSE2());
 	printf("AltiVec: %d\n", SDL_HasAltiVec());
+	printf("ARM SIMD: %d\n", SDL_HasARMSIMD());
+	printf("ARM NEON: %d\n", SDL_HasARMNEON());
 	return 0;
 }
 
