@@ -1,6 +1,6 @@
 /*
 	SDL - Simple DirectMedia Layer
-    Copyright (C) 1997-2004 Sam Lantinga
+    Copyright (C) 1997-2012 Sam Lantinga
 
 	This library is free software; you can redistribute it and/or
 	modify it under the terms of the GNU Library General Public
@@ -39,11 +39,7 @@
 /* The header was moved here in Mac OS X 10.1 */
 #include <Kernel/IOKit/hidsystem/IOHIDUsageTables.h>
 #endif
-#if MAC_OS_X_VERSION_MIN_REQUIRED == 1030
-#include "10.3.9-FIX/IOHIDLib.h"
-#else
 #include <IOKit/hid/IOHIDLib.h>
-#endif
 #include <IOKit/hid/IOHIDKeys.h>
 #include <CoreFoundation/CoreFoundation.h>
 #include <Carbon/Carbon.h> /* for NewPtrClear, DisposePtr */
@@ -186,7 +182,7 @@ static IOReturn HIDCreateOpenDeviceInterface (io_object_t hidDevice, recDevice *
 			plugInResult = (*ppPlugInInterface)->QueryInterface (ppPlugInInterface,
 								CFUUIDGetUUIDBytes (kIOHIDDeviceInterfaceID), (void *) &(pDevice->interface));
 			if (S_OK != plugInResult)
-				HIDReportErrorNum ("Couldnt query HID class device interface from plugInInterface", plugInResult);
+				HIDReportErrorNum ("Couldn't query HID class device interface from plugInInterface", plugInResult);
 			(*ppPlugInInterface)->Release (ppPlugInInterface);
 		}
 		else
@@ -344,6 +340,22 @@ static void HIDAddElement (CFTypeRef refElement, recDevice* pDevice)
 							headElement = &(pDevice->firstButton);
 						}
 						break;
+					case kHIDPage_Simulation:
+                        {
+                            switch (usage) /* look at usage to determine function */
+                            {
+                                case kHIDUsage_Sim_Rudder:
+                                case kHIDUsage_Sim_Throttle:
+                                    element = (recElement *) NewPtrClear (sizeof (recElement));
+                                    if (element)
+                                    {
+                                        pDevice->axes++;
+                                        headElement = &(pDevice->firstAxis);
+                                    }
+                                    break;
+                            }							
+                        }
+						break;
 					default:
 						break;
 				}
