@@ -1,6 +1,6 @@
 /*
     SDL - Simple DirectMedia Layer
-    Copyright (C) 1997-2006 Sam Lantinga
+    Copyright (C) 1997-2012 Sam Lantinga
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -50,6 +50,7 @@ static struct {
 	{ "SideWinder Precision Pro", 4, 1, 0 },
 	{ "SideWinder 3D Pro", 4, 1, 0 },
 	{ "Microsoft SideWinder 3D Pro", 4, 1, 0 },
+	{ "Microsoft SideWinder Precision Pro", 4, 1, 0 },
 	{ "Microsoft SideWinder Dual Strike USB version 1.0", 2, 1, 0 },
 	{ "WingMan Interceptor", 3, 3, 0 },
 	{ "WingMan Extreme Digital 3D", 4, 1, 0 },
@@ -58,6 +59,9 @@ static struct {
 	{ "Saitek Saitek X45", 6, 1, 0 }
 };
 
+/* It looks like newer kernels have the logical mapping at the driver level */
+#define NO_LOGICAL_JOYSTICKS
+
 #ifndef NO_LOGICAL_JOYSTICKS
 
 /*
@@ -370,13 +374,14 @@ static void LogicalSuffix(int logicalno, char* namebuf, int len)
 
 #if SDL_INPUT_LINUXEV
 #define test_bit(nr, addr) \
-	(((1UL << ((nr) & 31)) & (((const unsigned int *) addr)[(nr) >> 5])) != 0)
+	(((1UL << ((nr) % (sizeof(long) * 8))) & ((addr)[(nr) / (sizeof(long) * 8)])) != 0)
+#define NBITS(x) ((((x)-1)/(sizeof(long) * 8))+1)
 
 static int EV_IsJoystick(int fd)
 {
-	unsigned long evbit[40];
-	unsigned long keybit[40];
-	unsigned long absbit[40];
+	unsigned long evbit[NBITS(EV_MAX)] = { 0 };
+	unsigned long keybit[NBITS(KEY_MAX)] = { 0 };
+	unsigned long absbit[NBITS(ABS_MAX)] = { 0 };
 
 	if ( (ioctl(fd, EVIOCGBIT(0, sizeof(evbit)), evbit) < 0) ||
 	     (ioctl(fd, EVIOCGBIT(EV_KEY, sizeof(keybit)), keybit) < 0) ||
@@ -412,21 +417,31 @@ int SDL_SYS_JoystickInit(void)
 
 	numjoysticks = 0;
 
-	/* First see if the user specified a joystick to use */
+	/* First see if the user specified one or more joysticks to use */
 	if ( SDL_getenv("SDL_JOYSTICK_DEVICE") != NULL ) {
-		SDL_strlcpy(path, SDL_getenv("SDL_JOYSTICK_DEVICE"), sizeof(path));
-		if ( stat(path, &sb) == 0 ) {
-			fd = open(path, O_RDONLY, 0);
-			if ( fd >= 0 ) {
-				/* Assume the user knows what they're doing. */
-				SDL_joylist[numjoysticks].fname = SDL_strdup(path);
-				if ( SDL_joylist[numjoysticks].fname ) {
-					dev_nums[numjoysticks] = sb.st_rdev;
-					++numjoysticks;
-				}
-				close(fd);
+		char *envcopy, *envpath, *delim;
+		envcopy = SDL_strdup(SDL_getenv("SDL_JOYSTICK_DEVICE"));
+		envpath = envcopy;
+		while ( envpath != NULL ) {
+			delim = SDL_strchr(envpath, ':');
+			if ( delim != NULL ) {
+				*delim++ = '\0';
 			}
+			if ( stat(envpath, &sb) == 0 ) {
+				fd = open(envpath, O_RDONLY, 0);
+				if ( fd >= 0 ) {
+					/* Assume the user knows what they're doing. */
+					SDL_joylist[numjoysticks].fname = SDL_strdup(envpath);
+					if ( SDL_joylist[numjoysticks].fname ) {
+						dev_nums[numjoysticks] = sb.st_rdev;
+						++numjoysticks;
+					}
+					close(fd);
+				}
+			}
+			envpath = delim;
 		}
+		SDL_free(envcopy);
 	}
 
 	for ( i=0; i<SDL_arraysize(joydev_pattern); ++i ) {
@@ -472,8 +487,7 @@ int SDL_SYS_JoystickInit(void)
 					dev_nums[numjoysticks] = sb.st_rdev;
 					++numjoysticks;
 				}
-			} else
-				break;
+			}
 		}
 
 #if SDL_INPUT_LINUXEV
@@ -567,7 +581,7 @@ static SDL_bool JS_ConfigJoystick(SDL_Joystick *joystick, int fd)
 {
 	SDL_bool handled;
 	unsigned char n;
-	int old_axes, tmp_naxes, tmp_nhats, tmp_nballs;
+	int tmp_naxes, tmp_nhats, tmp_nballs;
 	const char *name;
 	char *env, env_name[128];
 	int i;
@@ -587,7 +601,6 @@ static SDL_bool JS_ConfigJoystick(SDL_Joystick *joystick, int fd)
 	}
 
 	name = SDL_SYS_JoystickName(joystick->index);
-	old_axes = joystick->naxes;
 
 	/* Generic analog joystick support */
 	if ( SDL_strstr(name, "Analog") == name && SDL_strstr(name, "-hat") ) {
@@ -658,9 +671,9 @@ static SDL_bool JS_ConfigJoystick(SDL_Joystick *joystick, int fd)
 static SDL_bool EV_ConfigJoystick(SDL_Joystick *joystick, int fd)
 {
 	int i, t;
-	unsigned long keybit[40];
-	unsigned long absbit[40];
-	unsigned long relbit[40];
+	unsigned long keybit[NBITS(KEY_MAX)] = { 0 };
+	unsigned long absbit[NBITS(ABS_MAX)] = { 0 };
+	unsigned long relbit[NBITS(REL_MAX)] = { 0 };
 
 	/* See if this device uses the new unified event API */
 	if ( (ioctl(fd, EVIOCGBIT(EV_KEY, sizeof(keybit)), keybit) >= 0) &&
@@ -689,33 +702,33 @@ static SDL_bool EV_ConfigJoystick(SDL_Joystick *joystick, int fd)
 				++joystick->nbuttons;
 			}
 		}
-		for ( i=0; i<ABS_MAX; ++i ) {
+		for ( i=0; i<ABS_MISC; ++i ) {
 			/* Skip hats */
 			if ( i == ABS_HAT0X ) {
 				i = ABS_HAT3Y;
 				continue;
 			}
 			if ( test_bit(i, absbit) ) {
-				int values[5];
+				struct input_absinfo absinfo;
 
-				if ( ioctl(fd, EVIOCGABS(i), values) < 0 )
+				if ( ioctl(fd, EVIOCGABS(i), &absinfo) < 0 )
 					continue;
 #ifdef DEBUG_INPUT_EVENTS
 				printf("Joystick has absolute axis: %x\n", i);
 				printf("Values = { %d, %d, %d, %d, %d }\n",
-					values[0], values[1],
-					values[2], values[3], values[4]);
+					absinfo.value, absinfo.minimum,
+					absinfo.maximum, absinfo.fuzz, absinfo.flat);
 #endif /* DEBUG_INPUT_EVENTS */
 				joystick->hwdata->abs_map[i] = joystick->naxes;
-				if ( values[1] == values[2] ) {
+				if ( absinfo.minimum == absinfo.maximum ) {
 				    joystick->hwdata->abs_correct[i].used = 0;
 				} else {
 				    joystick->hwdata->abs_correct[i].used = 1;
 				    joystick->hwdata->abs_correct[i].coef[0] =
-					(values[2] + values[1]) / 2 - values[4];
+					(absinfo.maximum + absinfo.minimum) / 2 - absinfo.flat;
 				    joystick->hwdata->abs_correct[i].coef[1] =
-					(values[2] + values[1]) / 2 + values[4];
-				    t = ((values[2] - values[1]) / 2 - 2 * values[4]);
+					(absinfo.maximum + absinfo.minimum) / 2 + absinfo.flat;
+				    t = ((absinfo.maximum - absinfo.minimum) / 2 - 2 * absinfo.flat);
 				    if ( t != 0 ) {
 					joystick->hwdata->abs_correct[i].coef[2] = (1 << 29) / t;
 				    } else {
@@ -813,6 +826,9 @@ int SDL_SYS_JoystickOpen(SDL_Joystick *joystick)
 		return(-1);
 	}
 	SDL_memset(joystick->hwdata, 0, sizeof(*joystick->hwdata));
+#if SDL_INPUT_LINUXEV
+	SDL_memset(joystick->hwdata->abs_map, ABS_MAX, sizeof(*joystick->hwdata->abs_map)*ABS_MAX);
+#endif
 	joystick->hwdata->fd = fd;
 
 	/* Set the joystick to non-blocking read mode */
@@ -922,6 +938,10 @@ void HandleHat(SDL_Joystick *stick, Uint8 hat, int axis, int value)
 	SDL_logical_joydecl(SDL_Joystick *logicaljoy = NULL);
 	SDL_logical_joydecl(struct joystick_logical_mapping* hats = NULL);
 
+	if (stick->nhats <= hat) {
+		return;  /* whoops, that shouldn't happen! */
+	}
+
 	the_hat = &stick->hwdata->hats[hat];
 	if ( value < 0 ) {
 		value = 0;
@@ -960,6 +980,9 @@ void HandleHat(SDL_Joystick *stick, Uint8 hat, int axis, int value)
 static __inline__
 void HandleBall(SDL_Joystick *stick, Uint8 ball, int axis, int value)
 {
+	if ((stick->nballs <= ball) || (axis >= 2)) {
+		return;  /* whoops, that shouldn't happen! */
+	}
 	stick->hwdata->balls[ball].axis[axis] += value;
 }
 
@@ -1086,6 +1109,9 @@ static __inline__ void EV_HandleEvents(SDL_Joystick *joystick)
 				}
 				break;
 			    case EV_ABS:
+				if (code >= ABS_MISC) {
+					break;
+				}
 				switch (code) {
 				    case ABS_HAT0X:
 				    case ABS_HAT0Y:
@@ -1100,15 +1126,17 @@ static __inline__ void EV_HandleEvents(SDL_Joystick *joystick)
 							events[i].value);
 					break;
 				    default:
-					events[i].value = EV_AxisCorrect(joystick, code, events[i].value);
+					if (joystick->hwdata->abs_map[code] != ABS_MAX ) {
+					  events[i].value = EV_AxisCorrect(joystick, code, events[i].value);
 #ifndef NO_LOGICAL_JOYSTICKS
-					if (!LogicalJoystickAxis(joystick,
+					  if (!LogicalJoystickAxis(joystick,
 				           joystick->hwdata->abs_map[code],
 					   events[i].value))
 #endif
-					SDL_PrivateJoystickAxis(joystick,
+					  SDL_PrivateJoystickAxis(joystick,
 				           joystick->hwdata->abs_map[code],
 					   events[i].value);
+					}
 					break;
 				}
 				break;
@@ -1191,8 +1219,8 @@ void SDL_SYS_JoystickQuit(void)
 
 	for ( i=0; SDL_joylist[i].fname; ++i ) {
 		SDL_free(SDL_joylist[i].fname);
+		SDL_joylist[i].fname = NULL;
 	}
-	SDL_joylist[0].fname = NULL;
 }
 
 #endif /* SDL_JOYSTICK_LINUX */
