@@ -1,6 +1,6 @@
 /*
     SDL - Simple DirectMedia Layer
-    Copyright (C) 1997-2006 Sam Lantinga
+    Copyright (C) 1997-2012 Sam Lantinga
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -32,8 +32,9 @@
 
 #if SDL_ASSEMBLY_ROUTINES
 #  if defined(__GNUC__) && (defined(__i386__) || defined(__x86_64__))
-#    define MMX_ASMBLIT 1
-#    define GCC_ASMBLIT 1
+     /* forced MMX to 0...it breaks on most compilers now.  --ryan. */
+#    define MMX_ASMBLIT 0
+#    define GCC_ASMBLIT 0
 #  elif defined(_MSC_VER) && defined(_M_IX86)
 #    if (_MSC_VER <= 1200)  
 #      include <malloc.h>   
@@ -234,12 +235,12 @@ static void BlitRGBtoRGBSurfaceAlpha128MMX(SDL_BlitInfo *info)
 	Uint32 *dstp = (Uint32 *)info->d_pixels;
 	int dstskip = info->d_skip >> 2;
 	Uint32 dalpha = info->dst->Amask;
-	Uint8 load[8];
+	Uint64 load;
 
-	*(Uint64 *)load = 0x00fefefe00fefefeULL;/* alpha128 mask */
-	movq_m2r(*load, mm4); /* alpha128 mask -> mm4 */
-	*(Uint64 *)load = 0x0001010100010101ULL;/* !alpha128 mask */
-	movq_m2r(*load, mm3); /* !alpha128 mask -> mm3 */
+	load = 0x00fefefe00fefefeULL;/* alpha128 mask */
+	movq_m2r(load, mm4); /* alpha128 mask -> mm4 */
+	load = 0x0001010100010101ULL;/* !alpha128 mask */
+	movq_m2r(load, mm3); /* !alpha128 mask -> mm3 */
 	movd_m2r(dalpha, mm7); /* dst alpha mask */
 	punpckldq_r2r(mm7, mm7); /* dst alpha mask | dst alpha mask -> mm7 */
 	while(height--) {
@@ -861,7 +862,7 @@ static void Blit32to565PixelAlphaAltivec(SDL_BlitInfo *info)
                 dR = (dstpixel >> 8) & 0xf8; \
                 dG = (dstpixel >> 3) & 0xfc; \
                 dB = (dstpixel << 3) & 0xf8; \
-                ACCURATE_ALPHA_BLEND(sR, sG, sB, sA, dR, dG, dB); \
+                ALPHA_BLEND(sR, sG, sB, sA, dR, dG, dB); \
                 *((unsigned short *)dst) = ( \
                     ((dR & 0xf8) << 8) | ((dG & 0xfc) << 3) | (dB >> 3) \
                 ); \
@@ -994,7 +995,7 @@ static void Blit32to32SurfaceAlphaKeyAltivec(SDL_BlitInfo *info)
             if(sA && Pixel != ckey) { \
                 RGB_FROM_PIXEL(Pixel, srcfmt, sR, sG, sB); \
                 DISEMBLE_RGB(((Uint8 *)dstp), 4, dstfmt, Pixel, dR, dG, dB); \
-                ACCURATE_ALPHA_BLEND(sR, sG, sB, sA, dR, dG, dB); \
+                ALPHA_BLEND(sR, sG, sB, sA, dR, dG, dB); \
                 ASSEMBLE_RGBA(((Uint8 *)dstp), 4, dstfmt, dR, dG, dB, dA); \
             } \
             dstp++; \
@@ -1097,7 +1098,7 @@ static void Blit32to32PixelAlphaAltivec(SDL_BlitInfo *info)
             DISEMBLE_RGBA((Uint8 *)srcp, 4, srcfmt, Pixel, sR, sG, sB, sA); \
             if(sA) { \
               DISEMBLE_RGBA((Uint8 *)dstp, 4, dstfmt, Pixel, dR, dG, dB, dA); \
-              ACCURATE_ALPHA_BLEND(sR, sG, sB, sA, dR, dG, dB); \
+              ALPHA_BLEND(sR, sG, sB, sA, dR, dG, dB); \
               ASSEMBLE_RGBA((Uint8 *)dstp, 4, dstfmt, dR, dG, dB, dA); \
             } \
             ++srcp; \
@@ -1296,7 +1297,7 @@ static void Blit32to32SurfaceAlphaAltivec(SDL_BlitInfo *info)
             unsigned sR, sG, sB, dR, dG, dB; \
             DISEMBLE_RGB(((Uint8 *)srcp), 4, srcfmt, Pixel, sR, sG, sB); \
             DISEMBLE_RGB(((Uint8 *)dstp), 4, dstfmt, Pixel, dR, dG, dB); \
-            ACCURATE_ALPHA_BLEND(sR, sG, sB, sA, dR, dG, dB); \
+            ALPHA_BLEND(sR, sG, sB, sA, dR, dG, dB); \
             ASSEMBLE_RGBA(((Uint8 *)dstp), 4, dstfmt, dR, dG, dB, dA); \
             ++srcp; \
             ++dstp; \
@@ -1432,6 +1433,66 @@ static void BlitRGBtoRGBSurfaceAlphaAltivec(SDL_BlitInfo *info)
 #endif
 #endif /* SDL_ALTIVEC_BLITTERS */
 
+#if SDL_ARM_SIMD_BLITTERS
+void BlitARGBto565PixelAlphaARMSIMDAsm(int32_t w, int32_t h, uint16_t *dst, int32_t dst_stride, uint32_t *src, int32_t src_stride);
+
+static void BlitARGBto565PixelAlphaARMSIMD(SDL_BlitInfo *info)
+{
+	int32_t width = info->d_width;
+	int32_t height = info->d_height;
+	uint16_t *dstp = (uint16_t *)info->d_pixels;
+	int32_t dststride = width + (info->d_skip >> 1);
+	uint32_t *srcp = (uint32_t *)info->s_pixels;
+	int32_t srcstride = width + (info->s_skip >> 2);
+
+	BlitARGBto565PixelAlphaARMSIMDAsm(width, height, dstp, dststride, srcp, srcstride);
+}
+
+void BlitRGBtoRGBPixelAlphaARMSIMDAsm(int32_t w, int32_t h, uint32_t *dst, int32_t dst_stride, uint32_t *src, int32_t src_stride);
+
+static void BlitRGBtoRGBPixelAlphaARMSIMD(SDL_BlitInfo *info)
+{
+	int32_t width = info->d_width;
+	int32_t height = info->d_height;
+	uint32_t *dstp = (uint32_t *)info->d_pixels;
+	int32_t dststride = width + (info->d_skip >> 2);
+	uint32_t *srcp = (uint32_t *)info->s_pixels;
+	int32_t srcstride = width + (info->s_skip >> 2);
+
+	BlitRGBtoRGBPixelAlphaARMSIMDAsm(width, height, dstp, dststride, srcp, srcstride);
+}
+#endif
+
+#if SDL_ARM_NEON_BLITTERS
+void BlitARGBto565PixelAlphaARMNEONAsm(int32_t w, int32_t h, uint16_t *dst, int32_t dst_stride, uint32_t *src, int32_t src_stride);
+
+static void BlitARGBto565PixelAlphaARMNEON(SDL_BlitInfo *info)
+{
+    int32_t width = info->d_width;
+    int32_t height = info->d_height;
+    uint16_t *dstp = (uint16_t *)info->d_pixels;
+    int32_t dststride = width + (info->d_skip >> 1);
+    uint32_t *srcp = (uint32_t *)info->s_pixels;
+    int32_t srcstride = width + (info->s_skip >> 2);
+
+    BlitARGBto565PixelAlphaARMNEONAsm(width, height, dstp, dststride, srcp, srcstride);
+}
+
+void BlitRGBtoRGBPixelAlphaARMNEONAsm(int32_t w, int32_t h, uint32_t *dst, int32_t dst_stride, uint32_t *src, int32_t src_stride);
+
+static void BlitRGBtoRGBPixelAlphaARMNEON(SDL_BlitInfo *info)
+{
+	int32_t width = info->d_width;
+	int32_t height = info->d_height;
+	uint32_t *dstp = (uint32_t *)info->d_pixels;
+	int32_t dststride = width + (info->d_skip >> 2);
+	uint32_t *srcp = (uint32_t *)info->s_pixels;
+	int32_t srcstride = width + (info->s_skip >> 2);
+
+	BlitRGBtoRGBPixelAlphaARMNEONAsm(width, height, dstp, dststride, srcp, srcstride);
+}
+#endif
+
 /* fast RGB888->(A)RGB888 blending with surface alpha=128 special case */
 static void BlitRGBtoRGBSurfaceAlpha128(SDL_BlitInfo *info)
 {
@@ -1883,13 +1944,13 @@ static void Blit565to565SurfaceAlphaMMX(SDL_BlitInfo *info)
 		Uint16 *dstp = (Uint16 *)info->d_pixels;
 		int dstskip = info->d_skip >> 1;
 		Uint32 s, d;
-		Uint8 load[8];
+		Uint64 load;
 	  
 		alpha &= ~(1+2+4);		/* cut alpha to get the exact same behaviour */
-		*(Uint64 *)load = alpha;
+		load = alpha;
 		alpha >>= 3;		/* downscale alpha to 5 bits */
 
-		movq_m2r(*load, mm0); /* alpha(0000000A) -> mm0 */
+		movq_m2r(load, mm0); /* alpha(0000000A) -> mm0 */
 		punpcklwd_r2r(mm0, mm0); /* 00000A0A -> mm0 */
 		punpcklwd_r2r(mm0, mm0); /* 0A0A0A0A -> mm0 */
 		/* position alpha to allow for mullo and mulhi on diff channels
@@ -1897,10 +1958,10 @@ static void Blit565to565SurfaceAlphaMMX(SDL_BlitInfo *info)
 		psllq_i2r(3, mm0);
 	  
 		/* Setup the 565 color channel masks */
-		*(Uint64 *)load = 0x07E007E007E007E0ULL;
-		movq_m2r(*load, mm4); /* MASKGREEN -> mm4 */
-		*(Uint64 *)load = 0x001F001F001F001FULL;
-		movq_m2r(*load, mm7); /* MASKBLUE -> mm7 */
+		load = 0x07E007E007E007E0ULL;
+		movq_m2r(load, mm4); /* MASKGREEN -> mm4 */
+		load = 0x001F001F001F001FULL;
+		movq_m2r(load, mm7); /* MASKBLUE -> mm7 */
 		while(height--) {
 			DUFFS_LOOP_QUATRO2(
 			{
@@ -2022,13 +2083,13 @@ static void Blit555to555SurfaceAlphaMMX(SDL_BlitInfo *info)
 		Uint16 *dstp = (Uint16 *)info->d_pixels;
 		int dstskip = info->d_skip >> 1;
 		Uint32 s, d;
-		Uint8 load[8];
+		Uint64 load;
 	  
 		alpha &= ~(1+2+4);		/* cut alpha to get the exact same behaviour */
-		*(Uint64 *)load = alpha;
+		load = alpha;
 		alpha >>= 3;		/* downscale alpha to 5 bits */
 
-		movq_m2r(*load, mm0); /* alpha(0000000A) -> mm0 */
+		movq_m2r(load, mm0); /* alpha(0000000A) -> mm0 */
 		punpcklwd_r2r(mm0, mm0); /* 00000A0A -> mm0 */
 		punpcklwd_r2r(mm0, mm0); /* 0A0A0A0A -> mm0 */
 		/* position alpha to allow for mullo and mulhi on diff channels
@@ -2036,10 +2097,10 @@ static void Blit555to555SurfaceAlphaMMX(SDL_BlitInfo *info)
 		psllq_i2r(3, mm0);
 
 		/* Setup the 555 color channel masks */
-		*(Uint64 *)load = 0x03E003E003E003E0ULL;
-		movq_m2r(*load, mm4); /* MASKGREEN -> mm4 */
-		*(Uint64 *)load = 0x001F001F001F001FULL;
-		movq_m2r(*load, mm7); /* MASKBLUE -> mm7 */
+		load = 0x03E003E003E003E0ULL;
+		movq_m2r(load, mm4); /* MASKGREEN -> mm4 */
+		load = 0x001F001F001F001FULL;
+		movq_m2r(load, mm7); /* MASKBLUE -> mm7 */
 		while(height--) {
 			DUFFS_LOOP_QUATRO2(
 			{
@@ -2814,6 +2875,22 @@ SDL_loblit SDL_CalculateAlphaBlit(SDL_Surface *surface, int blit_index)
 	   df->Bmask == 0x1f && SDL_HasAltiVec())
             return Blit32to565PixelAlphaAltivec;
         else
+#endif
+#if SDL_ARM_NEON_BLITTERS || SDL_ARM_SIMD_BLITTERS
+		if(sf->BytesPerPixel == 4 && sf->Amask == 0xff000000
+		   && sf->Gmask == 0xff00 && df->Gmask == 0x7e0
+		   && ((sf->Rmask == 0xff && df->Rmask == 0x1f)
+		   || (sf->Bmask == 0xff && df->Bmask == 0x1f)))
+		{
+#if SDL_ARM_NEON_BLITTERS
+		    if(SDL_HasARMNEON())
+		        return BlitARGBto565PixelAlphaARMNEON;
+#endif
+#if SDL_ARM_SIMD_BLITTERS
+		    if(SDL_HasARMSIMD())
+		        return BlitARGBto565PixelAlphaARMSIMD;
+#endif
+		}
 #endif
 	    if(sf->BytesPerPixel == 4 && sf->Amask == 0xff000000
 	       && sf->Gmask == 0xff00
@@ -2851,6 +2928,14 @@ SDL_loblit SDL_CalculateAlphaBlit(SDL_Surface *surface, int blit_index)
 			if(!(surface->map->dst->flags & SDL_HWSURFACE)
 				&& SDL_HasAltiVec())
 				return BlitRGBtoRGBPixelAlphaAltivec;
+#endif
+#if SDL_ARM_NEON_BLITTERS
+			if (SDL_HasARMNEON())
+				return BlitRGBtoRGBPixelAlphaARMNEON;
+#endif
+#if SDL_ARM_SIMD_BLITTERS
+			if (SDL_HasARMSIMD())
+				return BlitRGBtoRGBPixelAlphaARMSIMD;
 #endif
 			return BlitRGBtoRGBPixelAlpha;
 		}
