@@ -1,6 +1,6 @@
 /*
     SDL - Simple DirectMedia Layer
-    Copyright (C) 1997-2006 Sam Lantinga
+    Copyright (C) 1997-2012 Sam Lantinga
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -24,7 +24,7 @@
 /*
  * GEM SDL video driver implementation
  * inspired from the Dummy SDL driver
- * 
+ *
  * Patrice Mandin
  * and work from
  * Olivier Landemarre, Johan Klockars, Xavier Joubert, Claude Attard
@@ -32,25 +32,41 @@
 
 #include <gem.h>
 
+#include "SDL_timer.h"
 #include "../../events/SDL_sysevents.h"
 #include "../../events/SDL_events_c.h"
 #include "SDL_gemvideo.h"
 #include "SDL_gemevents_c.h"
+#include "SDL_gemmouse_c.h"
 #include "../ataricommon/SDL_atarikeys.h"	/* for keyboard scancodes */
 #include "../ataricommon/SDL_atarievents_c.h"
 #include "../ataricommon/SDL_xbiosevents_c.h"
 #include "../ataricommon/SDL_ataridevmouse_c.h"
 
+/* Duration after which we consider key released */
+
+#define KEY_PRESS_DURATION 100
+
+#define MSG_SDL_ID	(('S'<<8)|'D')
+
 /* Variables */
 
 static unsigned char gem_currentkeyboard[ATARIBIOS_MAXKEYS];
 static unsigned char gem_previouskeyboard[ATARIBIOS_MAXKEYS];
+static Uint32 keyboard_ticks[ATARIBIOS_MAXKEYS];
+
+static short prevmx=0,prevmy=0,prevmb=0;
+static short dummy_msgbuf[8] = {MSG_SDL_ID,0,0,0, 0,0,0,0};
 
 /* Functions prototypes */
 
-static int do_messages(_THIS, short *message);
-static void do_keyboard(short kc, short ks);
-static void do_mouse(_THIS, short mx, short my, short mb, short ks);
+static int do_messages(_THIS, short *message, short latest_msg_id);
+static void do_keyboard(short kc, Uint32 tick);
+static void do_keyboard_special(short ks, Uint32 tick);
+static void do_mouse_motion(_THIS, short mx, short my);
+static void do_mouse_buttons(_THIS, short mb);
+static int mouse_in_work_area(int winhandle, short mx, short my);
+static void clearKeyboardState(Uint32 tick);
 
 /* Functions */
 
@@ -58,6 +74,7 @@ void GEM_InitOSKeymap(_THIS)
 {
 	SDL_memset(gem_currentkeyboard, 0, sizeof(gem_currentkeyboard));
 	SDL_memset(gem_previouskeyboard, 0, sizeof(gem_previouskeyboard));
+	SDL_memset(keyboard_ticks, 0, sizeof(keyboard_ticks));
 
 	/* Mouse init */
 	GEM_mouse_relative = SDL_FALSE;
@@ -67,74 +84,76 @@ void GEM_InitOSKeymap(_THIS)
 
 void GEM_PumpEvents(_THIS)
 {
-	short mousex, mousey, mouseb, dummy;
-	short kstate, prevkc, prevks;
-	int i;
-	SDL_keysym	keysym;
+	short prevkc=0, mousex, mousey, mouseb, kstate;
+	int i, quit = 0;
+	SDL_keysym keysym;
+	Uint32 cur_tick;
+	static Uint32 prev_now = 0, prev_msg = 0;
+	static short latest_msg_id = 0;
 
-	SDL_memset(gem_currentkeyboard,0,sizeof(gem_currentkeyboard));
-	prevkc = prevks = 0;
-	
-	for (;;)
-	{
-		int quit, resultat, event_mask, mouse_event;
-		short buffer[8], kc;
-		short x2,y2,w2,h2;
+	cur_tick = SDL_GetTicks();
+	if (prev_now == cur_tick) {
+		return;
+	}
+	prev_now = cur_tick;
 
-		quit =
-			mouse_event =
-			x2=y2=w2=h2 = 0;
+	SDL_AtariMint_BackgroundTasks();
+	clearKeyboardState(cur_tick);
 
-		event_mask = MU_MESAG|MU_TIMER|MU_KEYBD;
-		if (!GEM_fullscreen && (GEM_handle>=0)) {
-			wind_get (GEM_handle, WF_WORKXYWH, &x2, &y2, &w2, &h2);
-			event_mask |= MU_M1;
-			mouse_event = ( (SDL_GetAppState() & SDL_APPMOUSEFOCUS)
-				== SDL_APPMOUSEFOCUS) ? MO_LEAVE : MO_ENTER;
+	if (prev_msg) {
+		/* Wait at least 20ms before each event processing loop */
+		if (cur_tick-prev_msg < 20) {
+			return;
 		}
+	}
+	prev_msg = cur_tick;
+
+	dummy_msgbuf[1] = ++latest_msg_id;
+	if (appl_write(GEM_ap_id, sizeof(dummy_msgbuf), dummy_msgbuf) == 0) {
+		/* If it fails, wait for previous id */
+		--latest_msg_id;
+	}
+
+	while (!quit) {
+		int resultat;
+		short buffer[8], kc, dummy;
 
 		resultat = evnt_multi(
-			event_mask,
+			MU_MESAG|MU_TIMER|MU_KEYBD,
 			0,0,0,
-			mouse_event,x2,y2,w2,h2,
+			0,0,0,0,0,
 			0,0,0,0,0,
 			buffer,
-			10,
+			1000,
 			&dummy,&dummy,&dummy,&kstate,&kc,&dummy
 		);
 
 		/* Message event ? */
 		if (resultat & MU_MESAG)
-			quit = do_messages(this, buffer);
+			quit = do_messages(this, buffer, latest_msg_id);
 
 		/* Keyboard event ? */
 		if (resultat & MU_KEYBD) {
-			if ((prevkc != kc) || (prevks != kstate)) {
-				do_keyboard(kc,kstate);
-			} else {
-				/* Avoid looping, if repeating same key */
-				break;
+			do_keyboard_special(kstate, cur_tick);
+			if (prevkc != kc) {
+				do_keyboard(kc, cur_tick);
+				prevkc = kc;
 			}
 		}
 
-		/* Mouse entering/leaving window */
-		if (resultat & MU_M1) {
-			if (this->input_grab == SDL_GRAB_OFF) {
-				/* Switch mouse focus state */
-				SDL_PrivateAppActive((mouse_event == MO_ENTER),
-					SDL_APPMOUSEFOCUS);
-			}
-			GEM_CheckMouseMode(this);
+		/* Timer event ? Just used as a safeguard */
+		if (resultat & MU_TIMER) {
+			quit = 1;
 		}
-
-		/* Timer event ? */
-		if ((resultat & MU_TIMER) || quit)
-			break;
 	}
 
-	/* Update mouse */
+	GEM_CheckMouseMode(this);
+
+	/* Update mouse state */
 	graf_mkstate(&mousex, &mousey, &mouseb, &kstate);
-	do_mouse(this, mousex, mousey, mouseb, kstate);
+	do_keyboard_special(kstate, cur_tick);
+	do_mouse_motion(this, mousex, mousey);
+	do_mouse_buttons(this, mouseb);
 
 	/* Now generate keyboard events */
 	for (i=0; i<ATARIBIOS_MAXKEYS; i++) {
@@ -142,7 +161,7 @@ void GEM_PumpEvents(_THIS)
 		if (gem_currentkeyboard[i] && !gem_previouskeyboard[i])
 			SDL_PrivateKeyboard(SDL_PRESSED,
 				SDL_Atari_TranslateKey(i, &keysym, SDL_TRUE));
-			
+
 		/* Key unpressed ? */
 		if (gem_previouskeyboard[i] && !gem_currentkeyboard[i])
 			SDL_PrivateKeyboard(SDL_RELEASED,
@@ -166,39 +185,39 @@ void GEM_PumpEvents(_THIS)
 	}
 }
 
-static int do_messages(_THIS, short *message)
+static int do_messages(_THIS, short *message, short latest_msg_id)
 {
-	int quit, posted, check_mouse_mode;
-	short x2,y2,w2,h2;
+	int quit, update_work_area, sdl_resize;
 
-	quit = check_mouse_mode = 0;
+	quit = update_work_area = sdl_resize = 0;
 	switch (message[0]) {
+		case MSG_SDL_ID:
+			quit=(message[1] == latest_msg_id);
+			break;
 		case WM_CLOSED:
-		case AP_TERM:    
-			posted = SDL_PrivateQuit();
+		case AP_TERM:
+			SDL_PrivateQuit();
 			quit=1;
 			break;
 		case WM_MOVED:
 			wind_set(message[3],WF_CURRXYWH,message[4],message[5],message[6],message[7]);
+			update_work_area = 1;
 			break;
 		case WM_TOPPED:
 			wind_set(message[3],WF_TOP,message[4],0,0,0);
 			/* Continue with TOP event processing */
 		case WM_ONTOP:
 			SDL_PrivateAppActive(1, SDL_APPINPUTFOCUS);
-			if (VDI_setpalette) {
-				VDI_setpalette(this, VDI_curpalette);
-			}
-			check_mouse_mode = 1;
+			VDI_setpalette(this, VDI_curpalette);
 			break;
 		case WM_REDRAW:
 			if (!GEM_lock_redraw) {
-				GEM_wind_redraw(this, message[3],&message[4]);
+				GEM_wind_redraw(this, message[3], &message[4], SDL_FALSE);
 			}
 			break;
 		case WM_ICONIFY:
 		case WM_ALLICONIFY:
-			wind_set(message[3],WF_ICONIFY,message[4],message[5],message[6],message[7]);
+			wind_set (message[3],WF_ICONIFY,message[4],message[5],message[6],message[7]);
 			/* If we're active, make ourselves inactive */
 			if ( SDL_GetAppState() & SDL_APPACTIVE ) {
 				/* Send an internal deactivate event */
@@ -206,30 +225,37 @@ static int do_messages(_THIS, short *message)
 			}
 			/* Update window title */
 			if (GEM_refresh_name && GEM_icon_name) {
-				wind_set(GEM_handle,WF_NAME,(short)(((unsigned long)GEM_icon_name)>>16),(short)(((unsigned long)GEM_icon_name) & 0xffff),0,0);
+				wind_set(GEM_handle,WF_NAME,
+					(short)(((unsigned long)GEM_icon_name)>>16),
+					(short)(((unsigned long)GEM_icon_name) & 0xffff),
+					0,0);
 				GEM_refresh_name = SDL_FALSE;
 			}
-			check_mouse_mode = 1;
+			GEM_iconified = SDL_TRUE;
+			update_work_area = 1;
 			break;
 		case WM_UNICONIFY:
-			wind_set(message[3],WF_UNICONIFY,message[4],message[5],message[6],message[7]);
+			wind_set (message[3],WF_UNICONIFY,message[4],message[5],message[6],message[7]);
 			/* If we're not active, make ourselves active */
 			if ( !(SDL_GetAppState() & SDL_APPACTIVE) ) {
 				/* Send an internal activate event */
 				SDL_PrivateAppActive(1, SDL_APPACTIVE);
 			}
 			if (GEM_refresh_name && GEM_title_name) {
-				wind_set(GEM_handle,WF_NAME,(short)(((unsigned long)GEM_title_name)>>16),(short)(((unsigned long)GEM_title_name) & 0xffff),0,0);
+				wind_set(GEM_handle,WF_NAME,
+					(short)(((unsigned long)GEM_title_name)>>16),
+					(short)(((unsigned long)GEM_title_name) & 0xffff),
+					0,0);
 				GEM_refresh_name = SDL_FALSE;
 			}
-			check_mouse_mode = 1;
+			GEM_iconified = SDL_FALSE;
+			update_work_area = 1;
 			break;
 		case WM_SIZED:
 			wind_set (message[3], WF_CURRXYWH, message[4], message[5], message[6], message[7]);
-			wind_get (message[3], WF_WORKXYWH, &x2, &y2, &w2, &h2);
+			update_work_area = sdl_resize = 1;
 			GEM_win_fulled = SDL_FALSE;		/* Cancel maximized flag */
 			GEM_lock_redraw = SDL_TRUE;		/* Prevent redraw till buffers resized */
-			SDL_PrivateResize(w2, h2);
 			break;
 		case WM_FULLED:
 			{
@@ -246,9 +272,8 @@ static int do_messages(_THIS, short *message)
 					GEM_win_fulled = SDL_TRUE;
 				}
 				wind_set (message[3], WF_CURRXYWH, x, y, w, h);
-				wind_get (message[3], WF_WORKXYWH, &x2, &y2, &w2, &h2);
+				update_work_area = sdl_resize = 1;
 				GEM_lock_redraw = SDL_TRUE;		/* Prevent redraw till buffers resized */
-				SDL_PrivateResize(w2, h2);
 			}
 			break;
 		case WM_BOTTOMED:
@@ -256,115 +281,173 @@ static int do_messages(_THIS, short *message)
 			/* Continue with BOTTOM event processing */
 		case WM_UNTOPPED:
 			SDL_PrivateAppActive(0, SDL_APPINPUTFOCUS);
-			if (VDI_setpalette) {
-				VDI_setpalette(this, VDI_oldpalette);
-			}
-			check_mouse_mode = 1;
+			VDI_setpalette(this, VDI_oldpalette);
 			break;
 	}
 
-	if (check_mouse_mode) {
-		GEM_CheckMouseMode(this);
+	if (update_work_area) {
+		GEM_align_work_area(this, message[3], SDL_TRUE);
+
+		if (sdl_resize) {
+			SDL_PrivateResize(GEM_work_w, GEM_work_h);
+		}
 	}
-	
+
 	return quit;
 }
 
-static void do_keyboard(short kc, short ks)
+static void do_keyboard(short kc, Uint32 tick)
 {
 	int scancode;
 
 	if (kc) {
 		scancode=(kc>>8) & (ATARIBIOS_MAXKEYS-1);
 		gem_currentkeyboard[scancode]=0xFF;
+		keyboard_ticks[scancode]=tick;
 	}
+}
+
+static void do_keyboard_special(short ks, Uint32 tick)
+{
+	int scancode=0;
 
 	/* Read special keys */
 	if (ks & K_RSHIFT)
-		gem_currentkeyboard[SCANCODE_RIGHTSHIFT]=0xFF;
+		scancode=SCANCODE_RIGHTSHIFT;
 	if (ks & K_LSHIFT)
-		gem_currentkeyboard[SCANCODE_LEFTSHIFT]=0xFF;
+		scancode=SCANCODE_LEFTSHIFT;
 	if (ks & K_CTRL)
-		gem_currentkeyboard[SCANCODE_LEFTCONTROL]=0xFF;
+		scancode=SCANCODE_LEFTCONTROL;
 	if (ks & K_ALT)
-		gem_currentkeyboard[SCANCODE_LEFTALT]=0xFF;
+		scancode=SCANCODE_LEFTALT;
+
+	if (scancode) {
+		gem_currentkeyboard[scancode]=0xFF;
+		keyboard_ticks[scancode]=tick;
+	}
 }
 
-static void do_mouse(_THIS, short mx, short my, short mb, short ks)
+static void do_mouse_motion(_THIS, short mx, short my)
 {
-	static short prevmousex=0, prevmousey=0, prevmouseb=0;
 	short x2, y2, w2, h2;
 
+	if (this->input_grab == SDL_GRAB_OFF) {
+		/* Switch mouse focus state */
+		if (!GEM_fullscreen && (GEM_handle>=0)) {
+			SDL_PrivateAppActive(
+				mouse_in_work_area(GEM_handle, mx,my),
+				SDL_APPMOUSEFOCUS);
+		}
+	}
+	GEM_CheckMouseMode(this);
+
 	/* Don't return mouse events if out of window */
 	if ((SDL_GetAppState() & SDL_APPMOUSEFOCUS)==0) {
 		return;
 	}
 
+	/* Relative mouse motion ? */
+	if (GEM_mouse_relative) {
+		if (GEM_usedevmouse) {
+			SDL_AtariDevMouse_PostMouseEvents(this, SDL_FALSE);
+		} else {
+			SDL_AtariXbios_PostMouseEvents(this, SDL_FALSE);
+		}
+		return;
+	}
+
 	/* Retrieve window coords, and generate mouse events accordingly */
 	x2 = y2 = 0;
 	w2 = VDI_w;
 	h2 = VDI_h;
 	if ((!GEM_fullscreen) && (GEM_handle>=0)) {
-		wind_get (GEM_handle, WF_WORKXYWH, &x2, &y2, &w2, &h2);
-
-		/* Do not generate mouse button event if out of window working area */
-		if ((mx<x2) || (mx>=x2+w2) || (my<y2) || (my>=y2+h2)) {
-			mb=prevmouseb;
-		}
+		x2 = GEM_work_x;
+		y2 = GEM_work_y;
+		w2 = GEM_work_w;
+		h2 = GEM_work_h;
 	}
 
-	/* Mouse motion ? */
-	if (GEM_mouse_relative) {
-		if (GEM_usedevmouse) {
-			SDL_AtariDevMouse_PostMouseEvents(this, SDL_FALSE);
-		} else {
-			SDL_AtariXbios_PostMouseEvents(this, SDL_FALSE);
-		}
-	} else {
-		if ((prevmousex!=mx) || (prevmousey!=my)) {
-			int posx, posy;
+	if ((prevmx!=mx) || (prevmy!=my)) {
+		int posx, posy;
 
-			/* Give mouse position relative to window position */
-			posx = mx - x2;
-			if (posx<0) posx = 0;
-			if (posx>w2) posx = w2-1;
-			posy = my - y2;
-			if (posy<0) posy = 0;
-			if (posy>h2) posy = h2-1;
+		/* Give mouse position relative to window position */
+		posx = mx - x2;
+		if (posx<0) posx = 0;
+		if (posx>w2) posx = w2-1;
+		posy = my - y2;
+		if (posy<0) posy = 0;
+		if (posy>h2) posy = h2-1;
 
-			SDL_PrivateMouseMotion(0, 0, posx, posy);
-		}
-		prevmousex = mx;
-		prevmousey = my;
+		SDL_PrivateMouseMotion(0, 0, posx, posy);
 	}
 
-	/* Mouse button ? */
-	if (prevmouseb!=mb) {
-		int i;
-
-		for (i=0;i<2;i++) {
-			int curbutton, prevbutton;
-		
-			curbutton = mb & (1<<i);
-			prevbutton = prevmouseb & (1<<i);
-		
-			if (curbutton && !prevbutton) {
-				SDL_PrivateMouseButton(SDL_PRESSED, i+1, 0, 0);
+	prevmx = mx;
+	prevmy = my;
+}
+
+static void do_mouse_buttons(_THIS, short mb)
+{
+	int i;
+
+	/* Don't return mouse events if out of window */
+	if ((SDL_GetAppState() & SDL_APPMOUSEFOCUS)==0)
+		return;
+
+	if (prevmb==mb)
+		return;
+
+	for (i=0;i<3;i++) {
+		int curbutton, prevbutton;
+
+		curbutton = mb & (1<<i);
+		prevbutton = prevmb & (1<<i);
+
+		if (curbutton && !prevbutton) {
+			SDL_PrivateMouseButton(SDL_PRESSED, i+1, 0, 0);
+		}
+		if (!curbutton && prevbutton) {
+			SDL_PrivateMouseButton(SDL_RELEASED, i+1, 0, 0);
+		}
+	}
+
+	prevmb = mb;
+}
+
+/* Check if mouse in visible area of the window */
+static int mouse_in_work_area(int winhandle, short mx, short my)
+{
+	short todo[4];
+	short inside[4] = {mx,my,1,1};
+
+	/* Browse the rectangle list */
+	if (wind_get(winhandle, WF_FIRSTXYWH, &todo[0], &todo[1], &todo[2], &todo[3])!=0) {
+		while (todo[2] && todo[3]) {
+			if (rc_intersect((GRECT *)inside,(GRECT *)todo)) {
+				return 1;
 			}
-			if (!curbutton && prevbutton) {
-				SDL_PrivateMouseButton(SDL_RELEASED, i+1, 0, 0);
+
+			if (wind_get(winhandle, WF_NEXTXYWH, &todo[0], &todo[1], &todo[2], &todo[3])==0) {
+				break;
 			}
 		}
-		prevmouseb = mb;
+
 	}
+
+	return 0;
+}
 
-	/* Read special keys */
-	if (ks & K_RSHIFT)
-		gem_currentkeyboard[SCANCODE_RIGHTSHIFT]=0xFF;
-	if (ks & K_LSHIFT)
-		gem_currentkeyboard[SCANCODE_LEFTSHIFT]=0xFF;
-	if (ks & K_CTRL)
-		gem_currentkeyboard[SCANCODE_LEFTCONTROL]=0xFF;
-	if (ks & K_ALT)
-		gem_currentkeyboard[SCANCODE_LEFTALT]=0xFF;
+/* Clear key state for which we did not receive events for a while */
+
+static void clearKeyboardState(Uint32 tick)
+{
+	int i;
+
+	for (i=0; i<ATARIBIOS_MAXKEYS; i++) {
+		if (keyboard_ticks[i]) {
+			if (tick-keyboard_ticks[i] > KEY_PRESS_DURATION) {
+				gem_currentkeyboard[i]=0;
+				keyboard_ticks[i]=0;
+			}
+		}
+	}
 }
